#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/logging/AmpelLoggingError.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.12.2018
# Last Modified Date:  18.01.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

class AmpelLoggingError(Exception):
	pass
